@@ -21,4 +21,8 @@
 
 /*  Prototypes for raw.c  */
 void raw_rawping(int host, int msec);
-void raw_rawhost(int host, int addr);
+#ifdef ENABLE_IPV6
+void raw_rawhost(int host, struct in6_addr * addr);
+#else
+void raw_rawhost(int host, struct in_addr * addr);
+#endif
